-- Eval.e -------------------------------------------------------------------
-- Include file for parsing and evaluating a sequence containing a math
-- formula in text form.
-- --------------------------------------------------------------------------
-- Mangan_Fred@yahoo.com.au   (2010)

-- EvalV1 works very well, but confuses the '(' and ')' characters with
-- 40 and 41, respectively (and not surprisingly). I hope to solve this
-- version of the include file (Eval.e).

-- An attempted solution using sprintf() to convert numbers to strings
-- ended up in confusion and (apparent) failure of the conversion in some
-- cases.(integers < 256 return as characters).

-- Strategy: computational numbers must always be distinguishable from
-- operators, and braces. Instead of storing a number as an atom value we
-- will store it as a sequence of type TokenNumber = {'!',atom} where
-- atom is the actual value of the number. Computing functions must accept
-- and return numbers in this format. Functions (Sin, etc) will still
-- continue to be simple text-sequences.
-- -------------------------------------------------------------------------
--
--	22/09/2014 5:38 PM; Added the following functions to this file, as
--	commonly required standard math methods. They use EvaluateFormula:
--	 	Integration:
--	 		Trapezoidal Rule		- IntegrateTrap()
--			
--		Differentiation:			- Derivative()
--			Lim ( f(x+h)-f(x) )/h
--			as h ---> 0
--
--	These make return in a format similar to that for EvaluateFormula().
--	
--	Platform Euphoria.
--

include wildcard.e        -- upper()
include get.e             -- value()

    constant MATHOPS = "^/*+-",        -- in precedence order
             BRACES = ")(",            -- separator braces
             MATHFUNCS = {"PI", "SQR","SIN","COS","TAN","ASIN","ACOS","ATAN",
                                "SEC","COSEC","COT","ABS","EXP","LOG","LN"},
             EXPNO = 2.7182818,        -- exponential number
             WHITESPACE = " \t\\r\n",
             MINUSSUB = {{'!', -1},'*'}        -- substitute for '-' in some cases

    -- displayable text describing rules briefly
    global
    constant MATH_RULES = {
                     "FORMULA RULES",
                     " ",
                     "1. Enter formula as you would for an ",
                     "   old-style hand calculator.",
                     " ",
                     "2. Basic math operators are:",
                     "      ^ raise to a power;",
                     "      * multiply;",
                     "      / divide;",
                     "      + add;",
                     "      - subtract",
                     " ",
                     "3. Supported maths functions:",
                     "     Sin()      Cos()    Tan()",
                     "     CoSec()    Sec()    CoTan()",
                     "     Asin()     Acos()   Atan()",
                     "     Abs()      Exp()    Ln( )",
                     "     Log( )     Sqr()",
                     " ",
                     "4. Functions take precedence over other",
                     "   math ops on the same level, e.g.:",
                     "             Sin(x)^2 ",
                     "   is interpreted as (Sin(x))^2",
                     "   NOT as             Sin(x^2).",
                     "   Use braces to avoid ambiguity.",
                     " ",
                     "5. Enter angular quantities as radian.",
                     " ",
                     "6. Pi (= 3.14159..) can be entered as",
                     "   'PI' (or 'pi' or 'Pi').",
                     " ",
                     "7. Numbers may be entered in exponent ",
                     "   form, (eg, 1.2E-3 ---> 1.2 x 10^-3)",
                     "   as well as in integer and decimal",
                     "   form.",
                     " ",
                     "8. Spaces may be included in formula to ",
                     "   aid readability.",
                     " ",
                     "9. Interpretation is insensitive to ",
                     "   case. Thus, these are all equivalent:",
                     "      SIN  sin  Sin",
                     " ",
                     "10. In dynamic applications, you can ",
                     "   have a variable 'X' or 'x' in your ",
                     "   formula.",
                     "   A value must be supplied for 'X'",
                     "   or 'x' which will be substituted",
                     "   in for the variable.",
                     " ",
                     "11. Evaluation will not trap overflow",
                     "   or underflow, but will trap most",
                     "   other errors including division by 0",
                     "   error, attempt to log a negative ",
                     "   number and assorted formula syntax ",
                     "   errors.",
                     " ",
                     "DISCLAIMER -- The Author can accept no",
                     " responsibility for malfunction of this",
                     " software. Use at your own risk.",
                     " -------------------------------------",
                     "          (c) Fred Mangan (2010)",
                     " "
                  }

    integer ErrorCode       -- flags errors by number
            ErrorCode = 0   -- no error condition

             -- text to match errors detected in ErrorCode
    constant ERRMSG = {"Unmatched braces in formula",
                       "A ')' brace precedes any matching '(' brace",
                       "Division by 0",
                       "Formula is empty",
                       "Unresolvable math operators",
                       "Invalid character(s) in formula",
                       "Use of unrecognised function",
                       "Division by 0 in a function",
                       "Attempt to get log of n <= 0",
                       "Unknown math operator or missing arg.",
                       "Alien character(s) in formula",
                       "Faulty exponential notation",
                       "Attempt to find root of negative number.",
                       "No term(s) in a (sub-)expression",
                       "Attempt to raise negative number to non-integer power.",
                       "Undefined internal computational error",
                       "May not use PI when calculating in degree mode."
                      },

             BADCHARS = "';,?:[]{}|!@#$&_="  -- must not occur in formula


    -- Most errors are fatal to further calculation with the same formula.
    -- But a calling program might wish to attempt recovery from the
    -- following (e.g., in plotting programs):
    global constant DIV0 = 3,     -- division by 0 error
                    FNDIV0 = 8,   -- division by 0 in Function error
                    NEGLOG = 9,   -- attempt to log negative number
                    NEGROOT = 13  -- attempt to root negative number

    global sequence DumpedParse        -- fetch this to view or dump parsing
                    DumpedParse = {}   -- start it empty

    sequence MasterFormula      -- holds formula (for reentry)
    MasterFormula = ""          -- start it null

    sequence XIndex-- global to this page, to index 'X' in parsed formula

-- Define Types -----------------------------------------------------------

type char(object c)
    -- Test for c a 'character'
        if integer(c) then
            return (c > 0) and (c < 256)
        else
            return 0
        end if
end type -- char -----------------------------------------------------------

type DigityToken(char digit)
    -- Chars that may appear in an input
    -- number, excluding any exponent part of it.
    return find(digit, "0123456789.")
end type -- DigityToken ----------------------------------------------------

type TokenNumber(object s)
    -- Defines form of a token number - see introductory notes
        if sequence(s) then
            if length(s) = 2 and s[1] = '!' and atom(s[2]) then
                return 1
            end if
        end if

        return 0
end type -- TokenNumber ----------------------------------------------------

global procedure DumpRules()
-- Prints rules crudely to DOS screen
    for i = 1 to length(MATH_RULES) do
        printf(1, "%s\n", {MATH_RULES[i]})
    end for
end procedure -- DumpRules -------------------------------------------------

global function RetRules()
-- Returns MATH_RULES converted to '\r\n' delimiters - For windows programs.
	sequence rules = ""
	
	for i = 1 to length( MATH_RULES ) do
		rules &= MATH_RULES[i] & "\r\n"
	end for
	
	return rules
end function -- RetRules ---------------------------------------------------

-- Parsing Section ---------------------------------------------------------

procedure CheckBraces(sequence FormulaText)
-- Checks there is same number of '(' braces as of ')' braces
    integer chr, balance

    balance = 0
    for i = 1 to length(FormulaText) do
        chr = FormulaText[i]
        if chr = BRACES[1] then
            balance += 1
        elsif chr = BRACES[2] then
            balance -= 1
        end if
    end for

    if balance != 0 then
        ErrorCode = 1                                -- unmatched braces
    end if

end procedure -- CheckBraces -----------------------------------------------

function XtractDigits(sequence string)
-- Extracts DigityToken characters from start of FormulaText into a
-- 'numeric' sequence, which is returned
    sequence number
    integer dig_point

    number = ""
    dig_point = 1
    while DigityToken(string[dig_point]) do
        number &= string[dig_point]
        dig_point += 1
    end while

    return number
end function -- XtractDigits -----------------------------------------------

function ExtractNumber(sequence String)
-- This extracts a number from text String, starting at position 1.
-- This upgrade will extract numbers such as "1.0e-3" as well as
-- numbers in integer and decimal notation.
-- The extracted number is returned as type TokenNumber
-- Entry:        String = text sequence, starting with 1st digit of number
-- Return:   {lth, TokenNumber)
--                        lth = offset length for caller
--                        TokenNumber = number in form {'!', number}
    sequence link_text, extension, v
    integer p, lth
    object number

    -- get the front end of the number
    number = XtractDigits(String)

    p = length(number)

    if p > 0 then
        -- have a number, is it in exponent form?
        if String[p+1] = 'E' then
            -- maybe we have a second part
            p += 2
            link_text = "E"
            -- this could be followed by a +, - or a DigitToken
            if String[p] = '+' or String[p] = '-' then
                -- it's a sign, collect it
                link_text &= String[p]
                p += 1
            elsif DigityToken(String[p]) then
                -- its a DigityToken, we're ready to collect it so
                -- do nothing here
            else
                -- it's not an exponential form
                -- just prepare to continue
                p = 0
            end if

            if p > 0 then
                -- collect the DigityTokens for the e-portion
                extension = XtractDigits(String[p..$])
                -- if we got any
                if length(extension) > 0 and length(extension) < 4
                                                                                 and not find('.',extension) then
                    -- complete the e-type number by combining components
                    number &= link_text & extension
                else
                    -- we got no Digity tokens, so
                    ErrorCode = 12        -- flag bad exponent format
                end if
            end if

        end if

    end if

    -- measure offset for caller
    lth = length(number)

    -- the number is text: convert to value
    v = value(number)
    if v[1] = GET_SUCCESS then
        number = v[2]
    else
        -- bad number
        number = 0    -- default
    end if

    return {lth,{'!', number}}
end function -- ExtractNumber ----------------------------------------------

function ExtractFunc(sequence String)
-- Extracts characters of a function name into func if present in String.
-- Entry:   FormulaText = representation of formula (in whole or part)
--                        its first few chars are the putative name
-- Returns:  func       = verified function name, if found in MATHFUNCS
--                      = "" if not verified
    sequence func

    func = ""       -- default to no function

    for i = 1 to length(MATHFUNCS) do
        if i = 1 then
            -- 1st func should be in text as "PI", not "PI("
            if match(MATHFUNCS[i],String) = 1 then
                func = MATHFUNCS[i]
                exit
            end if
        else
            -- other funcs should be found in the text in the form: "FUNC("
            if match(MATHFUNCS[i] & '(',String) = 1 then
                -- got the function
                func = MATHFUNCS[i]
                exit
            end if
        end if
    end for

    return func
end function -- ExtractFunc ------------------------------------------------

procedure CheckLegalCharacters(sequence FormulaText)
-- Check for illegal characters in sequence FormulaText;
-- this check picks up only those defined in BADCHARS
    for i=1 to length(BADCHARS) do
        if find(BADCHARS[i],FormulaText) then
            ErrorCode = 6                 -- is illegal char
        end if
    end for
end procedure -- CheckLegalCharacters --------------------------------------

function FirstParse(sequence Formula)
-- Procedure does an initial parsing of (math text) Formula into
-- discrete elements in sequence parsed. The elements in parsed
-- will be either chars or sub-sequences.
-- Entry:     Formula        = putative math formula
-- Returns:   parsed         = sequence of parsed elements; or
--                           = {} if Formula submitted was empty
-- parsed elements are:
-- sub-sequences: representing Numbers or Functions
-- chars:         representing any of ^ + - * /
    sequence parsed, func, temp
    integer nibble_posn
    TokenNumber number

    parsed = {}                -- to accumulate the elements of parsing

    if length(Formula) > 0 then

    -- embrace entire formula and convert to upper case
    Formula = '(' & upper(Formula) & ')'

    -- verify legality of characters in Formula
    CheckLegalCharacters(Formula)

    -- this loop nibbles the expression Formula to extinction from left
    -- as it goes
    while length(Formula) do

        if DigityToken(Formula[1]) then
            -- try to extract a number to a sequence
            temp = ExtractNumber(Formula)        -- get the number
            number = temp[2]
            parsed = append(parsed, number)      -- add it to parsing
            nibble_posn = temp[1]+1              -- how much to advance
            if ErrorCode then
                -- get out
                exit
            end if

        elsif find(Formula[1],MATHOPS & BRACES & 'X') then
            -- have a math operator or brace or 'X' - treat it
            parsed = append(parsed, Formula[1])
            nibble_posn = 2
        else
            -- try for a function
            func = ExtractFunc(Formula)
            if length(func) > 0 then
                -- it is a function
                if equal (func,"PI") then
                    -- special case function; add its value
                    parsed = append(parsed,{'!',PI}) -- as TokenNumber
                else
                    -- one of the other functions; add it as a sequence
                    parsed = append(parsed,func)
                end if
                nibble_posn = length(func)+1

        elsif find(Formula[1],WHITESPACE) then
            -- dump white space
            nibble_posn = 2        -- just advance
        else
            -- its an illegal character
            ErrorCode = 11
            nibble_posn = 2
        end if
    end if

    -- nibble off front of Formula
    Formula = Formula[nibble_posn..$]
    end while

    else
        ErrorCode = 4        -- empty expression error
    end if

    return parsed
end function -- FirstParse -------------------------------------------------

-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

function DegreeToRadian(atom arg, integer DegreeFlag)
-- if degree flag set, does angle conversion degree ----> radian
    if DegreeFlag then
        arg *= PI/180
    end if
    return arg
end function -- DegreeToRadian ---------------------------------------------

function RadianToDegree(atom arg, integer DegreeFlag)
-- if degree flag set, does angle conversion radian ----> degree
    if DegreeFlag then
        arg *= 180/PI
    end if
    return arg
end function -- RadianToDegree ---------------------------------------------

function DealLeadingSign(integer inx,sequence parsed)
-- Deals with problem of a leading sign at the start of an expression.
-- Any '+' sign is dropped; any '-' sign is effectively replaced by "-1 *".
-- Entry:        inx        = index to '(' in parsed list.
--                   parsed = list of textual parsed elements.

    if atom(parsed[inx+1]) then
        if parsed[inx+1] = '-' then
            -- have a leading '-'; substitute for it
            parsed = parsed[1..inx] & MINUSSUB & parsed[inx+2..$]
        elsif parsed[inx+1] = '+' then
            -- got a leading '+'; dump it
            parsed = parsed[1..inx] & parsed[inx+2..$]
        end if
    end if

    return parsed
end function -- DealLeadingSign --------------------------------------------

function DealDoubleOps(integer op1,integer op2,integer i,sequence parsed)
-- User may have adjacent ops in expression; deal with this
-- Entry:   op1 = first operator
--          op2 = second operator
--          i   = pointer to element after 2nd operator
--                parsed  = the sequence being parsed
-- Returns: parsed, with problem 'fixed'
--          ErrorCode = 6 if could not resolve
    object successor
    sequence good_double_ops

    --                  op1 op2   op1 op2
    good_double_ops = {{'+','+'},{'+','-'},       -- allowed double ops in
                       {'-','+'},{'-','-'},       -- appropriate circumstances
                       {'*','+'},{'*','-'},
                       {'^','+'},{'^','-'},
                       {'/','+'},{'/','-'}
                      }

    if find({op1,op2},good_double_ops) then

        -- have potentially resolvable ops
        successor = parsed[i]
        if TokenNumber(successor) then
            -- successor is representation of a number; can fix this.
            if op2 = '-' or op2 = '+' then
                if op2 = '-' then
                    parsed[i][2] *= -1
                elsif op2 = '+' then
                    -- do nothing
                end if
                -- close up to dump op2
                parsed = parsed[1..i-2] & parsed[i..$]
            else
                ErrorCode = 5        -- Unresolvable math operators
            end if
        else
            -- successor is a func token; can possibly resolve this
            -- treat as for leading sign problem earlier
            if op2 = '-' then
                -- incorporate as MINUSSUB
                parsed = parsed[1..i-2] & MINUSSUB & parsed[i..$]
            elsif op2 = '+' then
                -- dump the + and close up
                parsed = parsed[1..i-2] & parsed[i..$]
            else
                ErrorCode = 5       -- Unresolvable math operators
            end if
        end if
    else
        ErrorCode = 5               -- Unresolvable math operators
    end if

    return parsed
end function -- DealDoubleOps ----------------------------------------------

function IndexXOccurences(sequence parsed)
-- Indexes occurences of 'X' in parsed, for future substitution
-- Uses XIndex (global this page) so can be reused

    XIndex = {}                 -- start it empty

    for i = 1 to length(parsed) do
        if char(parsed[i]) and parsed[i] = 'X' then
            XIndex = append(XIndex, i)        -- add index value
        end if
    end for

    return XIndex
end function -- IndexXOccurences -------------------------------------------

function RefineParse(sequence parsed, integer DegreeFlag)
-- Refines crude parsing done by FirstParse()
    object op1,op2
    integer i1

    if length(parsed) > 0 then

        -- treat any leading sign problem
        -- problem arises for "(-..." situations
        -- (while loop used here because parsed's length may vary (+ or -)
        --  during this process)
        i1 = 1
        while i1 <= length(parsed) do

            if atom(parsed[i1]) and parsed[i1] = '(' then
                -- have potential leading sign problem
                parsed = DealLeadingSign(i1,parsed)
            end if

            i1 += 1        -- advance

        end while

        -- treat other probable double-ops occurences;
        for i = 3 to length(parsed) do
            -- NECESSARY DUPLICATION?
            if i > length(parsed) then
                -- overshoot because parsed was shortened
                exit
            end if
            op1 = parsed[i-2]
            op2 = parsed[i-1]

            if char(op1) and char(op2) then  -- operators?
                if find(op1,MATHOPS) and find(op2,MATHOPS) then
                    -- they are operators; process them
                    parsed = DealDoubleOps(op1,op2,i,parsed)
                end if
            end if
        end for

        -- value pi should not occur in a formula if DegreeFlag is True
        if (find({'!',PI}, parsed) or find({'!',-PI}, parsed)) and DegreeFlag then
            ErrorCode = 17
        end if

        -- Index occurences of 'X' in parsed
        -- (any non-variable 'X's are now in sequences (eg "EXP"), so can do this)
        XIndex = IndexXOccurences(parsed)

    else
        ErrorCode = 4        -- empty expression
    end if

    return parsed
end function -- RefineParse ------------------------------------------------

function ParseCtl(sequence FormulaText, integer DegreeFlag)
-- Overall control of parsing process
-- Entry:   FormulaText = the formula as text
-- Returns: parsed      = parsed version of the text

    sequence parsed

    -- do initialling
    parsed = {}     -- an empty bucket
    ErrorCode = 0   -- no error so far

    CheckBraces(FormulaText)

    if not ErrorCode then

        parsed = FirstParse(FormulaText)         -- establish parsing sequence
        parsed = RefineParse(parsed, DegreeFlag) -- refine the parsing

    end if

    return parsed
end function -- ParseCtl ---------------------------------------------------

-- Evaluation Section ------------------------------------------------------

function TwoArgCalc(sequence pars_3)
-- Carries out calculation of a 2-argument elementary math operation
-- Entry:   pars_3 = a sequence of form {arg1, op, arg2}
--                   where arg1, arg2 are two TokenNumber's and
--                   op is one of '^','+', '-', '*', '/'
--Returns:  result = value of calculated result as TokenNumber
--                   may alter ErrorCode if a computational error condition
--                  arises.
    atom arg1,arg2, result
    integer op

    result = 0              -- default

    arg1 = pars_3[1][2]     -- extract info from a TokenNumber
    op = pars_3[2]          -- this one is a char (integer)
    arg2 = pars_3[3][2]     -- extract info from a TokenNumber

    if not find(op,MATHOPS) then
        ErrorCode = 10
    else
        -- carry out the operation
        if op = '^' then
            -- got a 'math function singularity error' here
            -- triggered by an 0^2 calculation
            if arg1 = 0 then
                -- duck 'math function singularity error'
                result = 0    -- this is the correct result
            elsif arg1 < 0 and not integer(arg2) then
                -- can't raise -ve number to non-integer power
                ErrorCode = 15
            else
                result = power(arg1,arg2)
            end if
        elsif op = '*' then
            result = arg1 * arg2
        elsif op = '/' then
            if arg2 = 0 then
                ErrorCode = 3        -- trap /0 error
            else
                result = arg1 / arg2
            end if
        elsif op = '+' then
            result = arg1 + arg2
        elsif op = '-' then
            result = arg1 - arg2
        end if

    end if

    return {'!', result}        -- as TokenNumber type
end function -- TwoArgCalc -------------------------------------------------

procedure VerifyOps(sequence sim_exp)
-- sim_exp is sequence containing a number, or alternating numbers and
-- operators. If sim_exp is correct in form,
--    * every 2nd element should be a math op, if there are any;
--    * sim_exp should contain an odd number of terms.
-- This prcedure checks that these conditions are met, and sets
-- ErrorCode if they are not.

    -- sim_exp should have an odd number of elements
    if integer(length(sim_exp)/2) then
        ErrorCode = 10            -- error: missing number or operator
    else
        -- check the operators
        for i = 2 to length(sim_exp) by 2 do
            if not find(sim_exp[i], MATHOPS) then
                ErrorCode = 10    -- missing operator
                exit
            end if
        end for
    end if

end procedure -- VerifyOps -------------------------------------------------

function EvaluateSimpleExpression(sequence simple_exp)
-- This function evaluates the numerical expression in simple_exp.
-- simple_exp is a simple expression of form
--      {'(',n[1], op[1], n[2], op[2], ... op[n-1], n[n],')'}
-- where n[i] are numbers of type TokenNumber and op[i] are standard math
-- operators: '^ * / + -'.
-- There is an indeterminate number of '(' at the start, with equivalent
-- number of ')' at the end of the sequence.
    atom p
    TokenNumber result

    -- strip off surrounding braces if any
    p = 1
    while p and length(simple_exp) > 1 do
        p = find('(', simple_exp)
            if p then
                simple_exp = simple_exp[2..$-1]
            end if
    end while

    result = {'!', 0}        -- dummy default

    if length(simple_exp) > 0 then
        -- check that ops are patterned properly
        VerifyOps(simple_exp)

        while length(simple_exp) > 1 and not ErrorCode do
            -- for each operator type, (in priority order)
            for i = 1 to length(MATHOPS) do
                p = 1
                while p and length(simple_exp) > 1 and not ErrorCode do
                    -- find the next operator
                    p = find(MATHOPS[i],simple_exp)
                    if p then
                        -- have operator;
                        -- evaluate triplet arg1, op, arg2 and save result
                        simple_exp [p-1] = TwoArgCalc(simple_exp[p-1..p+1])
                        -- and close up
                        simple_exp = simple_exp[1..p-1] & simple_exp[p+2..$]
                    end if
                end while
            end for
        end while

        -- fetch result
        if TokenNumber(simple_exp[1]) then
            result = simple_exp[1]
        else
            ErrorCode = 16    -- undefined computational error
        end if

    else
        ErrorCode = 14        -- no term(s) in a (sub-)expression
    end if

    return result        -- it is a TokenNumber
end function -- EvaluateSimpleExpression -----------------------------------

function ComputeFunc(TokenNumber f_arg, sequence func, integer DegreeFlag)
-- Computes one of the supported functions of MATHFUNCS
-- Entry:   f_arg  = an intermediate value by calculation,
--                   as a TokenNumber
--          func   = a member of MATHFUNCS, (already verified)
    atom result, arg

    arg = f_arg[2]    -- get the numeric argument

    result = 0        -- default

    if equal(func,"SIN") then
        arg = DegreeToRadian(arg, DegreeFlag)
        result = sin(arg)
    elsif equal(func,"SQR") then
        if arg < 0 then
            ErrorCode = 13                 -- no sqroot of negatives
        else
            result = power(arg,0.5)
        end if
    elsif equal(func,"COS") then
        arg = DegreeToRadian(arg, DegreeFlag)
        result = cos(arg)
    elsif equal(func,"TAN") then
        arg = DegreeToRadian(arg, DegreeFlag)
        if cos(arg) != 0 then
            result = tan(arg)
        else
            -- may never occur because
            -- our PI is less than true pi due to
            -- truncation.
            ErrorCode = 8        -- /0 in a function
        end if
    elsif equal(func,"ASIN") then
        result = arcsin(arg)
        result = RadianToDegree(result, DegreeFlag)

    elsif equal(func,"ACOS") then
        result = arccos(arg)
        result = RadianToDegree(result, DegreeFlag)
    elsif equal(func,"ATAN") then
        result = arctan(arg)
        result = RadianToDegree(result, DegreeFlag)
    elsif equal(func,"SEC") then
        arg = DegreeToRadian(arg, DegreeFlag)
        if cos(arg) != 0 then
            result = 1 / cos(arg)
        else
            -- may never occur because ...
            ErrorCode = 8                -- /0 error in a function
        end if
    elsif equal(func,"COSEC") then
        arg = DegreeToRadian(arg, DegreeFlag)
        if sin(arg) != 0 then
            result = 1 / sin(arg)
        else
            -- may never occur because ...
            ErrorCode = 8                -- /0 error in a function
        end if
    elsif equal(func,"COT") then
        arg = DegreeToRadian(arg, DegreeFlag)
        if sin(arg) != 0 then
            result = cos(arg) / sin(arg)
        else
            -- may never occur because ...
            ErrorCode = 8                -- /0 error in a function
        end if
    elsif equal(func,"ABS") then
        if arg < 0 then
            result = arg * -1
        else
            result = arg
        end if
    elsif equal(func,"EXP") then
        result = power(EXPNO, arg)
    elsif equal(func,"LOG") then
        if arg > 0 then
            result = log(arg) / 2.30259
        else
            ErrorCode = 9                -- no log of -ve number, or of 0
        end if
    elsif equal(func,"LN") then
        if arg > 0 then
            result = log(arg)
        else
            ErrorCode = 9                -- no log of -ve number, or of 0
        end if
    else
        -- ought never execute; -- an "unknown function" will
        -- have been interpreted as illegal characters.
        result = 1           -- default
        ErrorCode = 7        -- use of an unrecognized function
    end if

    return {'!', result} -- as TokenNumber
end function -- ComputeFunc ------------------------------------------------

function SubstituteX(sequence parsed, atom x)
-- Substitute x at indexed position(s) in parsed

    for i = 1 to length(XIndex) do
        parsed[XIndex[i]] = {'!', x}        -- substitute as TokenNumber
    end for

    return parsed
end function -- SubstituteX ------------------------------------------------

function EvaluateCtl(sequence parsed, atom x, integer DegreeFlag)
-- Main routine controlling evaluation of parsing parsed
-- Strategy: While there are "inner brace pairs" do:
--        1. Find an innermost pair.
--        2. Evaluate its contents.
--        3. Replace braced pair and contents by the evaluation.
--        4. Repeat steps 1-3 until
--        5. Done.
    atom result
    integer p1, p2
    object test_func

    result = 99999    -- dummy result

    parsed = SubstituteX(parsed, x)

    while length(parsed) > 1 do

        -- find start for 'innermost' pair by search for the
        -- last '(' in sequence parsed.
        p1 = 0
        for i = 1 to length(parsed) do        -- this is slow!
            if char(parsed[i]) and parsed[i] = '(' then
                p1 = i
            end if
        end for

        -- complete any bracing positions identification
        if p1 then
            p2 = find(')', parsed[p1+1..$])
            p2 += p1
        end if

        if p1 > 0 and p2 > p1 then
            -- evaluate expression bracketed by p1 and p2 and save it
            parsed[p1] = EvaluateSimpleExpression(parsed[p1..p2])
            -- close up
            parsed = parsed[1..p1] & parsed[p2+1..$]
            -- there may be a function associated with value inter_result;
            -- if so compute it.
            if p1 > 1 then
                -- there is room for a function sequence
                test_func = parsed[p1-1]
                if sequence(test_func) and find(test_func,MATHFUNCS) then
                    -- compute the function value and save it
                    parsed[p1-1] = ComputeFunc(parsed[p1], test_func, DegreeFlag)
                    -- and close up
                    parsed = parsed[1..p1-1] & parsed[p1+1..$]
                end if
            end if
        else
            exit
        end if
    end while

    if TokenNumber(parsed[1]) then
        result = parsed[1][2]
    end if

    return result -- (atom)
end function -- EvaluateCtl ------------------------------------------------

global function EvaluateFormula(sequence FormulaText,    -- math formula
                                atom x,                  -- value of x
                                integer DegreeFlag)      -- if using degrees
-- Main calling function for formula parsing and evaluation
-- Entry:    FormulaText = math formula in text form (1-line)
--                     x = value for x (obligatory, even if no x in formula)
-- Returns:  On success  {n, y};   n = 0, y = numerical result (atom)
--           On failure  {n, Msg}; n = ErrorCode, Msg = error message.
    sequence parsed
    object result

    ErrorCode = 0                        -- everything OK ... so far

    if not equal(FormulaText,MasterFormula) then
        -- must do a parsing
        parsed = ParseCtl(FormulaText, DegreeFlag)
        -- save the parsing to global for use below
        -- made this public, for caller to look up (MUST NOT CHANGE)
        DumpedParse = parsed
        -- and the formula, for the test in this 'if' group
        MasterFormula = FormulaText
    end if

    -- retrieve parsed; to recycle on multiple calls using same formula
    parsed = DumpedParse

    if ErrorCode then
        -- something wrong in parsing; return it
        return {ErrorCode, ERRMSG[ErrorCode]}
    else
        -- good parse; go on to evaluate and return it
        result = EvaluateCtl(parsed, x, DegreeFlag)
        if ErrorCode then
            -- something wrong in evaluating; return it
            return {ErrorCode, ERRMSG[ErrorCode]}
        else
            -- return successful result
            return {ErrorCode, result}
        end if
    end if

end function -- EvaluateFormula --------------------------------------------

-- -------------------------------------------------------------------------

	--integer true = 1, false = not true
				
global function IntegrateTrap
		( atom a, atom b, integer n, sequence formula, integer degree_flag)
-- Finds area between curve defined by formula and the x-axis using the
-- Trapezoidal rule for integration between x-limits a and b:

--   F(x) = (b-a)/(2*n)*( f(a) + 2.SUM( f(x)1 ... f(x)n-1 ) + f(b) )

-- Note: negative area increments are made positive, so function returns
--		 result as the gross physical area bounded.
-- 		 GET_SUCCESS = 0 on success of EvaluateFormula(), (confusing).

-- Returns:		{ 0, trapeze_area } 				on success; or
--				{ err_code, err_text }				on failure.
	atom x, step_x
	sequence fab = {0,0}, ab = {a,b}, temp = {0,0}, sum = {0,0}
	object coef
		
	if n < 2 then
		-- force error condition
		temp = {9999, "n must be integer > 1 in Trapezoidal rule!"}
		return temp
	end if
	
	if a>b then
		-- correct a remediable fault
		coef = {a,b}
		b = coef[1]
		a = coef[2]
	end if
	
	if temp[1] = GET_SUCCESS then
	
		-- calculate f(a) and f(b) into fab 
		-- assumes correct calculation occurs here 
		for i=1 to 2 do 
			fab[i] = EvaluateFormula(formula,ab[i],degree_flag)
			-- if numeric and < 0 then; not numeric if an error string
			if atom(fab[i][2]) and fab[i][2] < 0 then
				fab[i][2] *= -1			-- make absolute value
			end if
		end for
		
		-- initialize main summation
		step_x = (b-a)/n		-- amount to increment x on each loop step
		x = a+step_x			-- starting x for loop
		coef = step_x/2			-- multiplier needed to complete calculation
						
		for i = 1 to n-1 do
			
			-- if an error occurs, temp[1] will become non-zero
			temp = EvaluateFormula(formula,x,degree_flag)
			if temp[1] != GET_SUCCESS then
				exit				-- can't evaluate
			else
				-- continue calculation
				if temp[2]<0 then
					temp[2] *= -1	-- make absolute value 
				end if
				sum += temp			-- update sum
			end if
			x+=step_x				-- advance
			
		end for
		
		if temp[1] = GET_SUCCESS then
			
			-- good outcome; complete the calculation:
			sum[2] *= 2						-- double it
			sum = fab[1] + sum + fab[2]		-- add the f(a) and f(b) elements
			sum[2] *= coef					-- finish the calculation
			temp = sum
			
		end if

	end if
	
	return temp
end function -- IntegrateTrap ----------------------------------------------

global function Derivative( atom x, atom h, sequence formula, integer degree_flag )
-- Calculates derivative of textual formula, at domain value x.
-- Straight-forward application of the fundamental equation:

--		 Lim	f( x + h ) - f( x )
-- f'( x ) =	   -----------------
--		 h-->0			  h

-- Note: GET_SUCCESS has value 0 for ok
	atom deriv
	atom error_ptr = 0			-- if calc error, will point at its source
	sequence error_text
	
	sequence terms = { EvaluateFormula(formula, (x+h), degree_flag),
					   EvaluateFormula(formula,  x,    degree_flag)
					 }

	for i=1 to 2 do
		if terms[i][1]!= GET_SUCCESS then
			error_ptr = i
			error_text = terms[i][2]	-- terms[i][2] has error text
		end if
	end for

	if not error_ptr then
		
		-- ok - calculate and return result
		deriv = ( terms[1][2] - terms[2][2] ) / h
		return { GET_SUCCESS, deriv }
		
	else
		
		-- return error condition
		return {error_ptr, error_text}
		
	end if
		
end function -- Derivative -------------------------------------------------



