-- Bustout.e ----------------------------------------------------------------
--  "Bustout" of various polygons.

-- These "bustout_" functions return poly-line sequences (which may be multi-
-- line) representing a variety of 2D figures:

-- _sincos():	 a Sine or a Cosine curve; as polyline
-- _rectangle(): a rectangle; as polygon
-- _arc():		 arc of a circle; as a polyline
-- _circle():	 circles, polygons or ellipses; as a polygon
-- _arc_slice(): e.g., like arc(), ends joined to centre; as polygon
-- _star():		 a "star" shape; as polygon
-- _rock():		 an "irregular" (!) circle; as polygon
-- _lisajoux():  produces a lisajoux figure;
-- _can_cursor():XORable cursor; compound; created canonically.


-- --------------------------------------------------------------------------
-- FredMangan6@gmail.com


	global constant PI = 3.141592654
	
type point(sequence s)
-- defines what a euphoria point is:
    return length(s) = 2 and atom(s[1]) and atom (s[2])
end type    -- point

global function bustout_line(point start, point stop, integer n_points = 1)
-- bustout a straight line from start to stop containing n_points
    sequence delta = {stop[1] - start[1],stop[2] - start[2]}/n_points
    sequence new_line = {}
    sequence this_point = start
    integer  point_count = 0

    while point_count <= n_points do
        new_line = append(new_line, this_point)
        this_point += delta
        point_count += 1
    end while

    return new_line
end function    -- bustout_line --------------------------------------------

global function bustout_jag_line(point start, point stop,
								integer n_points=2, atom zig=1)
    sequence new_line = {}
	atom theta, dx, dy

    if n_points < 3 then
		puts(1, "bustout_jag_line requires you nominate at least 3 points\n\n.")
	else
		new_line = bustout_line(start, stop, n_points)
		dx = new_line[2][1] - new_line[1][1]
		dy = new_line[2][2] - new_line[1][2]
		if dx != 0 and dy != 0 then
			theta = arctan(dy/dx)
		end if
		for i = 2 to length(new_line) by 2 do
			if dy = 0 then
				new_line[i][2] += zig
			elsif dx = 0 then
				new_line[i][1] += zig
			else
				new_line[i] += { -zig*cos(theta), zig*sin(theta)}
			end if

		end for
	end if

	return new_line
end function -- bustout_jag_line -------------------------------------------

global function bustout_perturb(sequence pline, atom perturb_fac)
-- Returns a perturbation of pline; the perturbation is proportional to the
-- distance between pline[1] and pline[2]	
	
	if length(pline) > 1 then
		
		for i=1 to length(pline) do
			for j=1 to 2 do
				pline[i][j] += perturb_fac * (rand(2)-1)  
			end for
		end for
		
	end if
	
	return pline
end function -- bustout_perturb --------------------------------------------

global function bustout_sincos
		    (integer nsteps, atom amp, atom x1, atom x2, atom n, atom phi, integer cosflag)
-- Generate sine curve, y = amp * sin( n*x+phi ), or a corresponding cosine curve if cosflag is set.
-- Entry:   nsteps  = number of steps in generating curve
--          amp     = amplitude of curve
--          x1      = xvalue to start at
--          x2      = x-value to end at (a < b)
--			n		= multiplier for x in trig func.
--          cosflag = 0: sine curve generated, else cosine curve
-- 1/11/2015 5:31 PM MODIFIED by introduction of n and phi.
    sequence sc
    atom temp
    
    if x1 > x2 then
		temp = x1    x1 = x2   x2 = temp
    end if
    
    sc = {}
    if nsteps > 2 then

      for x = x1 to x2 by (x2 - x1)/nsteps do
		if cosflag then
	    	sc = append(sc, {x, amp * cos( n*x+phi )})
		else
	    	sc = append(sc, {x, amp * sin( n*x+phi )})
		end if
      end for

    end if
    
    return sc
end function    -- bustout_sincos ------------------------------------------

global function bustout_rectangle(sequence diag) -- don't define diag here
-- Given diag(onal), returns the 5 vertices of the rectangle it defines.
-- This includes the closing point. - diag may slope up or down
    sequence rect = repeat(0, 5)

    -- create the rectangle
    rect = repeat(0, 5)
    rect[1] = diag[1]
    rect[2] = {diag[1][1], diag[2][2]}
    rect[3] = diag[2]
    rect[4] = {diag[2][1], diag[1][2]}
	rect[5] = rect[1]					-- close
	
    return rect
end function    -- bustout_rectangle ---------------------------------------

global function bustout_arc(atom n, object r, atom theta, atom phi)
-- Bustouts an arc of radius r, centered on {0,0}
-- The arc subtends theta, starts at angle phi and consists of n elements.
-- If r is sequence {a, b} the arc will be a part of an ellipse
-- Returns the arc as a sequence; it is not a polygon!
    sequence arc, errmsg
    atom a, b

    if atom(r) then
    	a = r    b = r
    elsif length(r) = 2 then
    	a = r[1] b = r[2]
    else
    	-- warning to programmer via DOS:
    	errmsg = "WARNING: Error in bustout_arc():\n" &
    			 "r must be an atom or 2-element sequence {r1,r2}."
    	puts(1, errmsg)
    	-- try to go on ... may crash
    	a = r[1] b = r[2]

    end if

    arc = {}
    for angle = phi to phi+theta by theta/n do
		arc = append(arc, {a * cos(angle), b * sin(angle)})
    end for
    return arc
end function    -- bustout_arc ---------------------------------------------

global function bustout_circle(integer num, point centre, object r)
-- generates a (polygon) circle or ellipse, radius r, centered on
-- centre = {xc, yc}. The polygon has num sides.
-- If r = {a, b}, will be an ellipse.
    sequence circ

    circ = bustout_arc(num, r, 2*PI, 0)     -- circle centered on {0, 0}
    circ += repeat(centre, length(circ))    -- displace to its position

    return circ
end function    -- bustout_circle ------------------------------------------

global function bustout_annulus(integer n_sides, sequence r)
-- Returns an annulus centered on {0,0} having radii r[1] and r[2].
-- order of r[1] and r[2] does not matter.
	sequence annulus = {}
	atom x, y, dtheta = 2*PI / n_sides

	for theta=0 to 2*PI by dtheta do
		
		x = r[1]*cos(theta)					-- a point for r[1] ...
		y = r[1]*sin(theta)
		annulus = prepend(annulus, {x,y})	-- all these will be prepended
		
		x = r[2]*cos(theta)					-- a point for r[2] ...
		y = r[2]*sin(theta)
		annulus = append(annulus, {x,y})	-- all these will be appended
		
	end for

	-- close the annulus
	annulus = append(annulus, annulus[1])
	
	return annulus
end function -- bustout_annulus --------------------------------------------

global function bustout_arc_slice
			(atom n, object r, point centre, atom theta, atom phi)
-- Bustouts a pie-chart type of slice.
-- Arc has radius r, centered on centre, subtends theta, start angled at
-- phi, and has n elements.
-- If r = {a, b}, will be elliptical
    sequence slice
    slice = bustout_arc(n, r, theta, phi)
    slice = prepend(slice, {0,0})           -- add centre
    slice = append(slice, {0,0})
    slice += repeat(centre, length(slice))  -- translate to position
    return slice
end function -- bustout_arc_slice -------------------------------------------

global function bustout_star(atom r_major, point centre, integer num_points)
-- Returns a polygon 'star' of radius r_major about centre = {Xc, Yc}.
-- The star has num_points.
-- Returns {starr, radii}
    sequence starr
    atom delta_theta, r, r_minor, count, x, y, xc, yc
    delta_theta = PI / num_points
    xc = centre[1]
    yc = centre[2]
    starr = {}
    r_minor = r_major / 2.71828
    count = 0
    for theta = 0 to 2 * PI by delta_theta do
      if remainder(count, 2) then               -- changing 2 is interesting
	r = r_major
      else
	r = r_minor
      end if
      x = r * cos(theta) + xc
      y = r * sin(theta) + yc
      starr = append(starr, {x, y})
      count += 1
    end for
    starr = append(starr, starr[1])				-- close
    
    return {starr, {r_major, r_minor}}
end function    -- bustout_star --------------------------------------------

global function bustout_rock(integer num, point centre, atom mean_r)
-- Returns a sequence containing an "irregular" (!) circle,
-- radius mean_r, centered on centre = {xc, yc}. The polygon has num sides.
	sequence rock
	atom r,x,y

	rock = {}
    for angle = 0 to 2*PI by 2*PI/num do
    	-- compute working r for this step, randomly, restricting to
    	-- +/- 35% of mean_r.
    	if rand(2) = 1 then
    		r = (1-(rand(35)/100)) * mean_r -- smaller radius than mean_r
    	else
    		r = (1+(rand(35)/100)) * mean_r -- bigger radius than mean_r
    	end if
    	-- compute coords
    	x = r * cos(angle) + centre[1]
    	y = r * sin(angle) + centre[2]
    	-- and add to rock
    	rock = append(rock, {x,y})
    end for
    -- close
    rock[$] = rock[1]

    return rock
end function -- bustout_rock -----------------------------------------------

global function bustout_lisajoux(atom xparam, atom yparam)
-- Returns a sequence containing the points of a lisajoux figure.
-- The figure is created in a 1x1 area and will need scaling by caller
    sequence lisa_seq
    atom width, height, x, y, step, maxt
    width = 1/2
    height = 1/2
    -- Adjust next 2 variables if quality is bad:
    step = 0.05     -- if too big, may cut sharp corners, get jaggies
    maxt = 1000     -- if too small, bits of figure may go missing,
		    -- but pattern may be better!
    if xparam > 1 or yparam > 1 then
	maxt = 100          -- more interesting
    end if
    lisa_seq = {}
    for t = 0  to maxt by step do
	x = width  * sin(xparam * t) + width
	y = height * cos(yparam * t) + height
	lisa_seq = append(lisa_seq, {x, y})
    end for
    return lisa_seq
end function -- bustout_lisajoux -------------------------------------------

global function bustout_can_xhair(integer radius)
-- Bustouts a canonical XORable drawing cursor of radius 'radius'
-- This would involve machine coordinates
	sequence curs

	curs = {{{2,0},{radius+3,0}},{{-2,0},{-(radius+3),0}},
			{{0,2},{0,radius+3}},{{0,-2},{0,-(radius+3)}}
		   }

	-- add the targeting circle
	curs = append(curs,bustout_circle(50,{0,0},radius))

	return curs
end function -- bustout_can_xhair ------------------------------------------

global function get_centre_poly(sequence poly)
-- Entry: poly should represent a symmetrical polygon
-- Returns: centre of polygon, else {}
	sequence centre = {}
	sequence sum = {0,0}
	atom npoints = length(poly)
	
	for i=1 to npoints do
		sum += poly[i]
	end for
	
	if npoints then
		sum /= npoints
	end if
	
	return sum
end function -- centre_poly ------------------------------------------------

