-- SScale. e ****************************************************************
-- Graphics scaling routines, usable by DOS or WIN graphics programs.
-- Include this file with programs needing scaling of lines & polygons.
-- This replaces my older graphic scaling.

-- Usage: with EuWinGUI:
--    Include the file "SPlotEWG.ew" in your program; it will automatically
--    include this file.
--    "SPlotEWG.ew" can be extended with your own custom general routines.
-- --------------------------------------------------------------------------
-- FredMangan6@gmail.com

    global constant MIN = 1,         -- assorted indices
                    MAX = 2,
                    X   = 1,
                    Y   = 2
    
                             -- defaults
    sequence SFacs            SFacs = {0,0}  -- scaling factors,  {FacX, FacY}
    global sequence C_View    C_View = {}    -- for copy of view, see SetScale()
    global sequence C_Win     C_Win = {}     -- for copy of win,  see SetScale()
    
-- **************************************************************************

function VerifyViewWindow(sequence rise_diags, sequence r_id)
-- Checks that neither view nor win will produce a delta = 0.
-- Entry:    rise_diags = {view, win}
    sequence component
    integer success
-- --------------------------------------------------------------------------
    
    success = 1
    component = {"Viewport","Window"}
    
    for i = 1 to length(rise_diags) do
        
        if (rise_diags[i][MAX][X] - rise_diags[i][MIN][X]) <= 0 or
           (rise_diags[i][MAX][Y] - rise_diags[i][MIN][Y]) <= 0 then
        
            -- error, inform programmer
            printf(1,"SCALING ERROR: %s =\n",{component[i]})
            ? rise_diags[i]
            printf(1, "is not a rising diagonal, in %s().\n\n",{r_id})
            system("Pause",1)
            
            -- flag failure
            success = 0
            exit
            
        end if
        
    end for
    
    return success
end function -- VerifyViewWindow -------------------------------------------

function VerifyScaleFacs(sequence r_id) -- r-id = name of caller routine
-- Checks that each scale factor != 0 (an error condition)
    integer success
-- -------------------------------------------------------------------------
    
    if SFacs[1] = 0 or SFacs[2] = 0 then
        
        -- error, inform programmer
        puts(1, "Error: Incorrect Scale Factors are set:\n")
        ? SFacs
        printf(1, "Called from routine %s\n\n", {r_id})
        system("Pause",1)
        
        -- flag failure
        success = 0
        
    else
        
        -- flag success
        success = 1
        
    end if
    
    return success
end function -- VerifyScaleFacs --------------------------------------------

global procedure SetScale(sequence view, -- viewport rising diag
                          sequence win   -- "world" window rising diag
                         )
-- Sets up scaling for drawing. Call before using other scale routines.
-- Can recall any time to alter scaling for different outputs.
-- Entry:    view = { {VminX,VminY}, {VmaxX,VmaxY} } for viewport
--           win  = { {WminX,WminY}, {WmaxX,WmaxY} } for world window
-- ------------------------------------------------------------------------
    
    if VerifyViewWindow({view,win},"SetScale") then
        
        -- go ahead with scaling:
        SFacs = {0,0}                -- to hold x-factor and y-factor
        
        for i= X to Y do
            SFacs[i] = (view[MAX][i]-view[MIN][i]) / (win[MAX][i]-win[MIN][i])
        end for
        
        -- save these for later use
        C_View = view
        C_Win = win
        
    else
        
        -- failure; close program
        close(0)    -- BAD; it assumes a DOS program
        
    end if
    
end procedure -- SetScale --------------------------------------------------

function FlipY(atom y_coord)    -- the y-coordinate to be reflected.
-- "Flips" y-coord for drawing ... relative to viewport.
-- ------------------------------------------------------------------------
    return C_View[MIN][Y] + (C_View[MAX][Y] - y_coord)
end function -- FlipY ------------------------------------------------------

global function ScalePoly(sequence poly)    -- a polyline or polygon sequence
-- Returns scaled version of polyline in poly
-- Entry:    SetScale() must have been executed.
-- Returns:  poly = the scaled version of poly.
    atom coord_pt
-- ------------------------------------------------------------------------
    
    if VerifyScaleFacs("ScalePoly") then
        
        for i = 1 to length(poly) do
            
            for j = X to Y do
                
                coord_pt = poly[i][j]
                coord_pt = C_View[MIN][j]+(coord_pt-C_Win[MIN][j])* SFacs[j]
                
                if j = Y then
                    -- flip y-coord
                    coord_pt = FlipY(coord_pt)
                end if
                
                -- save back
                poly[i][j] = coord_pt    -- tolerate non-integer results
                
            end for
            
        end for
        
    else
        
        -- terminate
        close(0)    -- BAD; assumes DOS program
        
    end if

    return poly
end function -- ScalePoly --------------------------------------------------

global function BackScalePoly(sequence poly) -- a scaled polyline or polygon
-- Do reverse scaling of a polyline.
-- Needed to convert, e.g., mouse coords etc to real world coordinates.
-- Entry:    SetScale() must have been executed.
-- Returns:  poly = the unscaled (i.e., "world") version of poly.
-- Be aware of the digital nature of this process and that back-
-- scaling may be excessively inaccurate to tolerances of what you want and
-- expect from it.
    
    atom coord
-- ------------------------------------------------------------------------
    
    if VerifyScaleFacs("BackScalePoly") then
        
        for i = 1 to length(poly) do
            
            for j = X to Y do
                
                coord = poly[i][j]
                
                if j = Y then
                    
                    -- undo flip y-coord
                    coord = FlipY(coord)
                    
                end if
                
                -- undo the scaling ... back to match window
                coord = C_Win[MIN][j]+(coord-C_View[MIN][j])/SFacs[j]
                
                -- save into poly
                poly[i][j] = coord
                
            end for
            
        end for
        
    else
        
        -- terminate
        close(0)    -- BAD; assumes DOS program
        
    end if
    
    return poly
end function -- BackScalePoly -----------------------------------------------

global function AskMachineCoord(sequence w_point) -- real world coord., {x,y}
-- Returns:  m_point    = {x',y'}, corresponding machine coord
    sequence m_point
    
    m_point = floor(ScalePoly( {w_point,w_point} ))
    
    return m_point[1]
end function -- AskMachineCoord --------------------------------------------

-- Clipping Routines -----------------------------------------------------
-- Although Picture-type controls do automatic clipping, that clipping often
-- leaves visible artifacts at the border of the picture area.
-- So these routines have been included to cope with this problem, should it
-- arise. Clipping brings a speed overhead.
-- ***************************************************************************

function w_intersect(integer edge, sequence line_seg)
-- Determine point {x, y}, if any, where line_seg {{x1, y1}, {x2, y2}} cuts
-- an edge.
-- Adjunct to w_clippoly()
     atom x, y, r
    
    if edge = 1 then
        
        r = (C_View[MAX][Y] - line_seg[1][Y]) / (line_seg[2][Y] - line_seg[1][Y])
        x = line_seg[1][X] + (line_seg[2][X] - line_seg[1][X]) * r
        y = C_View[MAX][Y]
        
     elsif edge = 2 then
        
        r = (C_View[MAX][X] - line_seg[1][1]) / (line_seg[2][1] - line_seg[1][1])
        y = line_seg[1][Y] + (line_seg[2][Y] - line_seg[1][Y]) * r
        x = C_View[MAX][X]
        
    elsif edge = 3 then
        
        r = (C_View[MIN][Y] - line_seg[1][Y]) / (line_seg[2][Y] - line_seg[1][Y])
        x = line_seg[1][X] + (line_seg[2][X] - line_seg[1][X]) * r
        y = C_View[MIN][Y]
        
    elsif edge = 4 then
        
        r = (C_View[MIN][X] - line_seg[1][X]) / (line_seg[2][X] - line_seg[1][X])
        y = line_seg[1][Y] + (line_seg[2][Y] - line_seg[1][Y]) * r
        x = C_View[MIN][X]
        
    end if
    
    return {x, y}
end function    -- w_intersect ----------------------------------------------

function w_visible(integer edge, sequence pt)
-- Sets v = 1 if point {x1, y1} is to the visible side of edge 1,2,3 or 4.
-- Called by w_clippoly.
  atom v1
  integer v

  if edge = 1 then
     v1 = (C_View[MAX][X] - C_View[MIN][X]) * (C_View[MAX][Y] - pt[2])
  elsif edge = 2 then
     v1 = -(C_View[MIN][Y] - C_View[MAX][Y]) * (C_View[MAX][X] - pt[1])
  elsif edge = 3 then
     v1 = (C_View[MIN][X] - C_View[MAX][X]) * (C_View[MIN][Y] - pt[2])
  elsif edge = 4 then
     v1 = -(C_View[MAX][Y] - C_View[MIN][Y]) * (C_View[MIN][X] - pt[1])
  end if
  
  if (v1 <= 0) then
    v = 0    -- on invisible side
  else
    v = 1    -- on visible side
  end if
  
  return v
end function    -- w_visible ------------------------------------------------

global function w_clippoly(sequence inpoly) -- polyline or polygon,
                                            -- machine coords
-- Using edge 1,2,3 or 4 clips a polygon of n vertices described in inpoly
-- to a polygon of nn vertices described in outpoly. This is done against
-- the (diagonal) view, assumed to be rising.
-- Adapted from "Teach yourself Computer Graphics", based on a recursive
-- algorithm attributed to Sutherland & Hodgman.
-- Note: a closed polygon passed in here will come out opened!
-- August 2007: greatly improved and debugged
-- September 2010: transferred to this SScale.e
    sequence temp, outpoly, s, p
    integer n, nn, count, vis1, vis2, vis3
    
    n = length(inpoly)
    temp = repeat({0, 0}, 2 * n)                -- create empty working seqs
    outpoly = temp                              -- same size as inpoly
    temp[1..n] = inpoly[1..n]                   -- copy inpoly to temp
    
    -- start the real work:
    for edge = 1 to 4 do                        -- for each boundary
      
      if n = 0 then                             -- out of coordinates, quit
        exit
      end if
      
      count = 1
      s = temp[n]                               --   s is last point
      
      for j = 1 to n do                         --   for each vertex
        
        p = temp[j]                             --     p is next point
        vis1 = w_visible(edge, p)
        if vis1 = 1 then                        --     p is visible
          vis2 = w_visible(edge, s)
          if vis2 = 1 then                      --       s is visible
            
            outpoly[count] = p
            count += 1
            
          else                                  --       s is invisible
            
            outpoly[count] = w_intersect(edge, {s, p})
            count += 1
            outpoly[count] = p
            count += 1
            
          end if -- vis2
          
        else                                --      p is invisible
          
          vis3 = w_visible(edge, s)
          if vis3 = 1 then
            
            outpoly[count] = w_intersect(edge, {s, p})
            count += 1
            
          end if -- vis3
          
        end if -- vis1
        
        s = p      -- move to next point
        
      end for -- j
      
      nn = count - 1
      n = nn
      
      for i = 1 to nn do                    -- assign temp with new polygon
        temp[i] = outpoly[i]
      end for -- i
      
    end for -- edge
    
    if nn = 0 then                          -- clipped out
        outpoly = {}                        --   return empty sequence
    else                                    -- there are points
        outpoly = outpoly[1..nn]            --   return relevant points
    end if
    
    return outpoly
end function    -- w_clippoly -----------------------------------------------

-- End of listing ***********************************************************

