-- File: Data.e

--------------------------------------------------------------------------------

-- Include files

--------------------------------------------------------------------------------
-- Include: IO.e
-- Location: main "include" folder
-- Called Routines: inherited
--------------------------------------------------------------------------------
include IO.e

--------------------------------------------------------------------------------
-- Include: ssl.e
-- Location: main "Include" folder
-- Called Routines: split
--------------------------------------------------------------------------------
include ssl.e as ssl

--------------------------------------------------------------------------------
-- Include: stats.e
-- Location: main "Include" folder
-- Called Routines: correlation
--------------------------------------------------------------------------------
include stats.e

--------------------------------------------------------------------------------

-- Local variables

--------------------------------------------------------------------------------
    object line
    sequence datum
    
--------------------------------------------------------------------------------

-- Data definition

--------------------------------------------------------------------------------
-- Class: Data (and prefix for methods)
--------------------------------------------------------------------------------
    Class("Data")
        Extends({"File"})
    
--------------------------------------------------------------------------------

-- Data Properties
    
--------------------------------------------------------------------------------
-- Property: string delimiter
-- Value:    delimiter used in the data records
--------------------------------------------------------------------------------
        Property("DELIMITER",STRING,",")

--------------------------------------------------------------------------------
-- Property: boolean labelFlag
-- Value:    set to TRUE if labels precede data
--------------------------------------------------------------------------------
        Property("LABEL_FLAG",BOOL,FALSE)
    
--------------------------------------------------------------------------------
-- Property: array_of_strings labels
-- Value:    variable labels
--------------------------------------------------------------------------------
        Property("LABELS",STRING_ARRAY,repeat("",1))
    
--------------------------------------------------------------------------------
-- Property: sequence matrix
-- Value:    data matrix
--------------------------------------------------------------------------------
        Property("MATRIX",SEQUENCE,{})

--------------------------------------------------------------------------------
-- Property: integer noVariables
-- Value:    number of variables to analyse
--------------------------------------------------------------------------------
        Property("NO_VARIABLES",INTEGER,0)
    
--------------------------------------------------------------------------------
-- Property: string runName
-- Value:    the title name associated with this run
--------------------------------------------------------------------------------
        Property("RUN_NAME",STRING,EMPTY_STRING)

--------------------------------------------------------------------------------
-- Property: boolean spssOutput
-- Value:    set to TRUE if SPSS-style output required
--------------------------------------------------------------------------------
        Property("SPSS_OUTPUT",BOOL,FALSE)
    
--------------------------------------------------------------------------------

-- Data Routines

--------------------------------------------------------------------------------
-- Method: LabelsRead
-- Usage:  MethodProc(<Object_Name>,"LabelsRead",{})
-- Purpose: reads the labels from the start of the file
-- Scope: local
--------------------------------------------------------------------------------
    procedure Data_LabelsRead_0(Instance self)
        string_array labels
        integer noVars
        labels = datum
        noVars = MethodFunc(self,"Get",{"NO_VARIABLES"})
        MethodProc(self,"Set",{"LABELS",labels})
        line = MethodFunc(self,"GetS",{})
    end procedure
    
--------------------------------------------------------------------------------

-- Define method: LabelsRead

--------------------------------------------------------------------------------
    Method("LabelsRead", routine_id("Data_LabelsRead_0"))

--------------------------------------------------------------------------------
-- Method: LabelsSet
-- Usage:  MethodProc(<Object_Name>,"LabelsSet",{})
-- Purpose: sets the labels to "..1", "..2", etc
-- Scope: local
--------------------------------------------------------------------------------
    procedure Data_LabelsSet_0(Instance self)
        string_array labels
        labels = repeat({},MethodFunc(self,"Get",{"NO_VARIABLES"}))
        for i = 1 to MethodFunc(self,"Get",{"NO_VARIABLES"}) do
            labels[i] = sprintf("%8d",i)
        end for
        MethodProc(self,"Set",{"LABELS",labels})
    end procedure
    
--------------------------------------------------------------------------------

-- Define method: LabelsSet

--------------------------------------------------------------------------------
    Method("LabelsSet", routine_id("Data_LabelsSet_0"))

--------------------------------------------------------------------------------
-- Method: MatrixOutput
-- Usage:  MethodProc(<Object_Name>,"MatrixOutput",{})
-- Purpose: outputs the correlation matrix to <fileName>.mat
-- Scope: local
--------------------------------------------------------------------------------
    procedure Data_MatrixOutput_0(Instance self)
        sequence matrixOutputFileName
        Instance matrixFile
        sequence matrix
        sequence path
        integer noVars
        path = ssl:split(MethodFunc(self,"Get",{"NAME"}),".")
        matrixOutputFileName = path[1] & ".mat"
        matrixFile = File(matrixOutputFileName,"w")
        matrix = MethodFunc(self,"Get",{"MATRIX"})
        noVars = MethodFunc(self,"Get",{"NO_VARIABLES"})
        for j = 1 to noVars do
            for k = 1 to noVars do
                MethodProc(matrixFile,"PrintFormat",
                           {"%010.7f",correlation(matrix[j],matrix[k])}
                          )
            end for
            MethodProc(matrixFile,"Puts",{"\n"})
        end for
        MethodProc(matrixFile,"Close",{})
    end procedure
    
--------------------------------------------------------------------------------

-- Define method: MatrixOutput

--------------------------------------------------------------------------------
    Method("MatrixOutput", routine_id("Data_MatrixOutput_0"))
    
--------------------------------------------------------------------------------
-- Method: GetParameters
-- Usage:  MethodProc(<Object_Name>,"GetParameters",{string runname,
--                                                   integer labflag,
--                                                   string delimiter,
--                                                   integer matflag
--                                                  }
--                    )
-- Purpose: establishes the parameters applying to this run of the program
-- Scope: local
-- Note: matrix initialised to all {}
--------------------------------------------------------------------------------
    procedure Data_GetParameters_4(Instance self,
                                   string runname,
                                   integer labflag,
                                   string delimiter,
                                   integer matflag
                                  )
        sequence matrix
        MethodProc(self,"Set",{"RUN_NAME",runname})
        MethodProc(self,"Set",{"LABEL_FLAG",labflag})
        MethodProc(self,"Set",{"DELIMITER",delimiter})
        MethodProc(self,"Set",{"SPSS_OUTPUT",matflag})
        -- calculate number of variables
        line = MethodFunc(self,"GetS",{})
        datum = ssl:split(line,MethodFunc(self,"Get",{"DELIMITER"}))
        MethodProc(self,"Set",{"NO_VARIABLES",length(datum)})
        matrix = repeat({},MethodFunc(self,"Get",{"NO_VARIABLES"}))
        MethodProc(self,"Set",{"MATRIX",matrix})
    end procedure
    
--------------------------------------------------------------------------------

-- Define method: GetParameters

--------------------------------------------------------------------------------
    Method("GetParameters", routine_id("Data_GetParameters_4"))

--------------------------------------------------------------------------------
-- Method: Report
-- Usage:  MethodProc(<Object_Name>,"Report",{})
-- Purpose: generates the "printed" output for this run
-- Scope: local
--------------------------------------------------------------------------------
    procedure Data_Report_0(Instance self)
        string outputFileName
        Instance resultsFile
        integer noRecords
        sequence path
        sequence matrix
        integer noVars
        string_array labels
        path = ssl:split(MethodFunc(self,"Get",{"NAME"}),".")
        outputFileName = path[1] & ".res"
        resultsFile = File(outputFileName,"w")
        matrix = MethodFunc(self,"Get",{"MATRIX"})
        noVars = MethodFunc(self,"Get",{"NO_VARIABLES"})
        noRecords = length(matrix[1])
        labels = MethodFunc(self,"Get",{"LABELS"})
        
        -- send to resultsFile
        MethodProc(resultsFile,"PrintFormat",{"** %s **\n\n",
                                           {MethodFunc(self,"Get",{"RUN_NAME"})}
                                             }
                  )
        MethodProc(resultsFile,"PrintFormat",{"Statistics for dataset: %s,",
                                              {MethodFunc(self,"Get",{"NAME"})}
                                             }
                  )
        MethodProc(resultsFile,"PrintFormat",{" based on %d records.\n\n",
                                              noRecords
                                             }
                  )
        MethodProc(resultsFile,"Puts",{"Field             Average      Variance"
                                       & "          SD\n\n"
                                      }
                  )
        for j = 1 to noVars do
            MethodProc(resultsFile,"PrintFormat",{"%-11s",
                                                  {labels[j]}
                                                 }
                      )
            MethodProc(resultsFile,"PrintFormat",{"%14.4f",
                                                  {average(matrix[j])}
                                                 }
                      )
            MethodProc(resultsFile,"PrintFormat",{"%14.4f",
                                                  {variance(matrix[j])}
                                                 }
                      )
            MethodProc(resultsFile,"PrintFormat",{"%14.4f",
                                                  {sqrt(variance(matrix[j]))}
                                                 }
                      )
            MethodProc(resultsFile,"Puts",{"\n"})
        end for
        MethodProc(resultsFile,"Puts",{"\nCovariances\n\n          "})
        for j = 1 to noVars do
            MethodProc(resultsFile,"PrintFormat",{"%14s",{labels[j]}})
        end for
        MethodProc(resultsFile,"Puts",{"\n"})
        for j = 1 to noVars do
            MethodProc(resultsFile,"PrintFormat",{"%-10s",{labels[j]}})
            for k = 1 to noVars do
                MethodProc(resultsFile,"PrintFormat",{"%14.4f",
                                                 covariance(matrix[j],matrix[k])
                                                     }
                          )
            end for
            MethodProc(resultsFile,"Puts",{"\n"})
        end for
        MethodProc(resultsFile,"Puts",{"\nCorrelations\n\n          "})
        for j = 1 to noVars do
            MethodProc(resultsFile,"PrintFormat",{"%10s",{labels[j]}})
        end for
        MethodProc(resultsFile,"Puts",{"\n"})
        for j = 1 to noVars do
            MethodProc(resultsFile,"PrintFormat",{"%-10s",{labels[j]}})
            for k = 1 to noVars do
                MethodProc(resultsFile,"PrintFormat",{"%14.4f",
                                                correlation(matrix[j],matrix[k])
                                                     }
                          )
            end for
            MethodProc(resultsFile,"Puts",{"\n"})
        end for
        MethodProc(resultsFile,"Close",{})
    end procedure
    
--------------------------------------------------------------------------------

-- Define method: Report

--------------------------------------------------------------------------------
    Method("Report",routine_id("Data_Report_0"))
    
--------------------------------------------------------------------------------
-- Method: Scan
-- Usage:  MethodProc(<Object_Name>,"Scan",{})
-- Purpose: scans the file
--          and processes the data into sequences within a sequence
-- Note: the first line of the file may contain the variable labels
-- Scope: local
--------------------------------------------------------------------------------
    procedure Data_Scan_0(Instance self)
        sequence partLine,
                 matrix
        matrix = repeat({},MethodFunc(self,"Get",{"NO_VARIABLES"}))
        if MethodFunc(self,"Get",{"LABEL_FLAG"}) then
            MethodProc(self,"LabelsRead",{})
        else
            MethodProc(self,"LabelsSet",{})
        end if
        while (sequence(line)) do
            datum = ssl:split(line,MethodFunc(self,"Get",{"DELIMITER"}))
            for i = 1 to MethodFunc(self,"Get",{"NO_VARIABLES"}) do
                partLine = value(datum[i])
                matrix[i] &= partLine[2]
            end for
            line = MethodFunc(self,"GetS",{})
        end while
        MethodProc(self,"Set",{"MATRIX",matrix})    
    end procedure

--------------------------------------------------------------------------------

-- Define method: Scan

--------------------------------------------------------------------------------
    Method("Scan",routine_id("Data_Scan_0"))
    
--------------------------------------------------------------------------------

--      end of Data definition
    
--------------------------------------------------------------------------------
    EndClass("Data")

--------------------------------------------------------------------------------
-- Function: Data
-- Usage Data()
-- Purpose: creates an Instance of Class Data
-- Scope: local
--------------------------------------------------------------------------------
    global function Data(string name,string mode)
    integer handle,
            self
        self = New("Data",{{"NAME",name}})
        handle = open(name,mode)
        if handle > 2 then
            MethodProc(self,"Set",{"HANDLE",handle})
        else
            printf(2,"**ERROR**\n*File '%s' cannot be opened*\n",{name})
        end if
        return self
    end function
