--------------------------------------------------------------------------------

-- Library EuCANOOP.e

--------------------------------------------------------------------------------
--/topic Site Header
--/info
--
--<div id="site_header">
--<div id="site_heading"><a href="http://sourceforge.net/projects/eucanoop/">EuCANOOP v2a.0.0</a></div>
--<div>OOP library for use with the <a href="http://www.rapideuphoria.com/">Euphoria Programming Language</a></div>
--</div>

--------------------------------------------------------------------------------
include misc.e

--/topic Introduction
--/H The EuCANOOP library.
--   /b<File: /i<EuCANOOP.e>> /n
--   /b<Version: /i<2.0.0>> /n
--   /b<Date: /i<28 October 2008>> /n
--   /b<Author: /i<C. A. Newbould>> /n
--   /b<Licence: /i<Free>> /n /n
--
-- OOP programming in Euphoria is not built into the language.
-- It is, however, quite simple to achieve and easy to apply.
-- The library in EuCANOOP provides a means of carrying out OOP
-- with just nine routines.
--/n /n
-- "Standard" Euphoria manipulates data structures such as
-- <i>object</i>s, <i>sequence</i>s, <i>atom</i>s,
-- <i>integer</i>s and any other type the user chooses to adopt
-- or define. It does the manipulation by means of routines, some
-- written in the program, others being in libraries.
-- Authors might be RDS, the user or, perhaps, other users.
--/n /n
-- EuCANOOP offers an exactly analogous form of processing, but
-- with just one data structure - the Object - and, at this stage
-- anyway, just two pre-written routines.
-- The difference is that the variables and routines are stored
-- within the data structure and manipulation takes place solely
-- within the Object.
--/n /n
-- You start by defining Classes, as prototypes of Objects.
-- After each definition is complete the details of that Class
-- are stored for future use. Once you have declared a Class you
-- can create Instances of it (Objects), setting values specific
-- to the Instance. Manipulation of an Object is possible because
-- it carries within its values not only its "variables" but also
-- pointers to routines (both <i>procedure</i>s and
-- <i>function</i>s) which operate on the Object's variables.
--/n /n
--This is all achieved using two standard Euphoria constructs:
--<ol>
--<li>the <i>sequence</i>
--<li>the function <i>routine_id</i>
--</ol>
-- The rest is what I once heard called "syntactical sugar" -
-- coding to provide stylistic simplification.
-- This is what the EuCANOOP library provides.

--------------------------------------------------------------------------------
--/topic Version Information
-- General version information.
--/n /n
-- This section documents changes in the EuCANOOP library across time.

--------------------------------------------------------------------------------
--/topic Version 2.0.0
--/parent Version Information
--<ul>
--/li Introduces type-checking for Properties & Methods (2008/09/01).
--/li Uses <a href=http://sourceforge.net/projects/eumakedoc/>makedoc</a>
-- for its documentation (2008/10/25).
--</ul>

--------------------------------------------------------------------------------
--/topic Classes and Objects
-- How to define Classes and Objects (instances).
-- You define a Class using the procedure /Class.
-- It takes one parameter:
--<ul>
--<li>a <i>sequence</i> containing the name you wish to associate
-- with that Class.
--</ul>
--/code
--Class("Dog")
--/endcode
--/n
-- If you wish the Class to inherit from another pre-defined
-- Class then use the procedure /Extends.
-- Multiple inheritance is supported.
-- /Extends takes one parameter:
--<ul>
--<li>an array of one or more <i>sequence</i>s
-- containing the name(s) of the Class(es).
--</ul>
--/code
--Extends({"Animal"})
--/endcode
--/n If you wish to add property values then you use the
-- procedure /Property. It takes three parameters:
--<ul>
--<li>a <i>sequence</i> containing the name you wish to associate
-- with the property.
--<li>an <i>integer</i> signifying the type of the variable
-- (ATOM for <i>atom</i>, SEQUENCE for <i>sequence</i>, etc).
--<li>an <i>object</i> containing the default value for the property.
-- This value should be of the correct type. If it isn't then a basic default
-- is set instead.
--</ul>
-- My convention is to use upper-case text for property names.
--/n /n
--/code
--Property("FEET",INTEGER,0)
--/endcode
--/n If you wish to add a method then use the procedure /Method.
-- It takes two parameters:
--<ul>
--<li>a <i>sequence</i> containing the name you wish to associate
-- with the method.
--<li>an <i>integer</i> carrying the return value from
-- <i>routine_id</i>.
--/n /n The parameter for that call should be the name of the routine you wish
-- to invoke when the method is called
-- (see later for how to code methods and how to invoke them).
-- The routine is either pre-defined, outside the Class definition, or
-- may be included within it.
--</ul>
-- My convention is to use title-case for method names.
--/n /n
--/code
--Method("Bark",routine_id("Dog_Bark"))
--/endcode
--/n Repeat Property and/or Method calls as required to define all the values
-- you need.
--/n /n If you wish to assign an inherited Method to a different routine then
-- you use the procedure /Override.
-- Just like /Method it takes two parameters:
--<ul>
--<li>a <i>sequence</i> containing the name of the inherited method.
--<li>an <i>integer</i> carrying the return value from
-- <i>routine_id</i> for the "new" routine.
--</ul>
-- A Class definition is only stored when the procedure
-- /EndClass is called. It takes one parameter:
--<ul>
--<li>a <i>sequence</i> containing the name you have already used to
-- define the Class.
--</ul>
--If you fail to include such a call, or key in the wrong name then the
-- Class will not be entered into the database of Classes.
--/n /n
--/code
--EndClass("Dog")
--/endcode
--/n I find it useful to use an indenting convention for Class declaration.
-- For example:
--/n /n
--/code
--Class("Dog")
--    Extends({"Animal"})
--    Property("BREED",SEQUENCE,"")
--    Method("Bark",routine_id("Dog_Bark"))
--EndClass("Dog")
--/endcode
--/n EuCANOOP's Objects are operational instances of Class paradigms.
-- Objects are created (in the image of their Class definition) using
-- the function /New. It takes two parameters:
--<ul>
--<li>a <i>sequence</i> containing the name of the Class from which
-- the Object derives.
--<li>an array of one or more arrays.
-- Each inner array contains two elements:
--<ul>
--<li>a <i>sequence</i> containing the name of the property whose
-- value you wish to set.
--<li>an <i>object</i> containing the value you wish to assign
-- to that property.
--(If the value is not of the correct type then the default setting for the
-- class is substituted.)
--</ul>
--</ul>
--You only need include such an array if you want to change from the default[s].
--/n /n
--/code
--Fido = New("Dog",{{"BREED","Mongrel"}})
--/endcode
--/n All Objects are created with two standard Methods:
-- /"Get" and /"Set".
--These Methods can be used to inspect or change the values of properties.
--/n /n Methods are called as follows:
--/n /n If the Method relates to a <i>procedure</i> then you use
--/MethodProc. It takes three parameters:
--<ul>
--<li>(<i>Instance</i>) name of the Object.
--<li>a <i>sequence</i> containing the name of the Method.
--<li>a <i>sequence</i> containing the parameters required by the
--procedure associated with the method - excluding the first (see later).
--</ul>
--/code
--MethodProc(SCREEN,"Puts",{"Hello, World!"})
--/endcode
--/n If the Method relates to a <i>function</i> then you use
--/MethodFunc. It takes three parameters:
--<ul>
--<li>(<i>Instance</i>) name of the Object.
--<li>a <i>sequence</i> containing the name of the Method.
--<li>a <i>sequence</i> containing the parameters required by the
--procedure associated with the method - excluding the first (see later).
--</ul>
--/code
--if MethodFunc(KEYBOARD,"GetC",{}) then
--end if
--/endcode
--/n Routines (qua Methods) must be defined as follows:
--/n /n The routine must have at least one parameter. This is an
--<i>Instance</i>, which is used to refer to the Object itself.
-- My convention is to use the name "self" for this parameter.
--The remaining parameters, if any, are the ones to be called by the Method.
--/n /n To avoid ambiguity my convention is to name routines in a consistent
--fashion. For example, if I intend to define a method called "What",
--for Class "Fred", with two method parameters (excluding "self") then I
--would name the routine <b>Fred_What_2</b>.
--/n /n Example:
--/n /n
--/code
--function Output_Puts_1(Instance self,sequence text)
--    return puts(self,text)
--end function
--/endcode

--------------------------------------------------------------------------------
--/topic Constructors
--/ Defining constructors.
--/n /n In some circumstances calling <code>New</code> is insufficient.
-- Moreover, some users with other OOP language experience actually want to
-- define Constructors. You can easily write conventional constructors by
-- combining a call to <code>New</code> with a relevant creator
-- routine. So, for example:</p>
--/code
--constant myFile = New("File",{{"NAME","Hello.ex"},{"MODE","r"}})
--MethodProc(myFile,"Open",{})
--/endcode
--/n could be replaced by
--/n /n
--/code
--constant myFile = File("Hello.ex","r")
--/endcode
--/n if the constructor <code>File</code> is defined as follows:
--/n /n
--/code
--global function File(sequence name,sequence mode)
--    integer obj,
--            handle
--    obj = New("File",{{"NAME",name},{"MODE",mode}})
--    handle = open(name,mode)
--    MethodProc(obj,"Set",{"HANDLE",handle})
--    return obj
--end function
--/endcode
--/n My convention is to use, as the creator function, a function with precisely
-- the same name as the Class.

--------------------------------------------------------------------------------
--/topic Destructors
-- Using destructors.
--/n /n Because Euphoria has built-in garbage collection, there is no real need
-- to define destructor routines. This even applies to operations such as
-- closing an already opened file - Euphoria automatically closes all files
-- when an application terminates.

--------------------------------------------------------------------------------

--Library specification

--------------------------------------------------------------------------------
--/topic Constants
-- Globally defined constants.
--/n /n The following constants are available for use.
--<ul>
--<li>FALSE = 0
--<li>TRUE = 1
--<li>EMPTY_STRING = ""
--</ul>

--------------------------------------------------------------------------------
global constant FALSE = 0,

                TRUE = 1,

                EMPTY_STRING = ""

--------------------------------------------------------------------------------
--/topic Service Routines
-- Global routines available for use.
--/n /n These are routines for type-checking of properties.

--------------------------------------------------------------------------------
--/topic Service Routines
--/func is_atom(object a)
-- Checks that <b>a</b> is an <i>atom</i>.

--------------------------------------------------------------------------------
global function is_atom(object a)
    if not atom(a) then
        return FALSE
    else
        return atom(a)
    end if
end function

--------------------------------------------------------------------------------
--/topic Service Routines
--/func is_instance(object a)
-- Checks that <b>a</b> is an <i>Instance</i>.

--------------------------------------------------------------------------------
global function is_instance(object i)
    integer inst
    inst = FALSE
    if not integer(i) then
        inst = FALSE
    else
        if i > -2 then
            inst = TRUE
        end if
    end if
    return inst
end function

--------------------------------------------------------------------------------
--/topic Service Routines
--/func is_integer(object a)
-- Checks that <b>a</b> is an <i>integer</i>.

--------------------------------------------------------------------------------
global function is_integer(object i)
    if not integer(i) then
        return FALSE
    else
        return integer(i)
    end if
end function

--------------------------------------------------------------------------------
--/topic Service Routines
--/func is_object(object a)
-- Checks that <b>a</b> is an <i>object</i>.

--------------------------------------------------------------------------------
global function is_object(object o)
    return TRUE
end function

--------------------------------------------------------------------------------
--/topic Service Routines
--/func is_sequence(object a)
-- Checks that <b>a</b> is a <i>sequence</i>.

--------------------------------------------------------------------------------
global function is_sequence(object s)
    if not sequence(s) then
        return FALSE
    else
        return sequence(s)
    end if
end function

--------------------------------------------------------------------------------
--/topic Types
-- Globally defined types.
--/n /n Objects are all declared to be of type <i>Instance</i>, which enables
-- the programmer to distinguish all his/her Object instances.
--/n /n
-- Just like Euphoria's variables, properties of Objects are typed. EuCANOOP,
-- however, uses its own type-checking routines, so property definitions use
-- EuCANOOP's own types - /i OBJECT, /i ATOM, /i INTEGER and /i SEQUENCE.
-- Additionally a /i METHOD type is defined to signal that a "Property" is,
-- in fact, a Method.
--/n /n
-- Type-checking in EuCANOOP takes place when a Class Property's default is
-- set, when an <i>Instance</i> is created with a non-default parameter
-- and when the built-in /Set Method is invoked. Unlike Euphoria, a type error
-- does not result in program termination. Although an error message is issued
-- if an assignment fails, the property's value is set, respectively, to a
-- "null" value for that type, the Class default or is left unchanged.

--------------------------------------------------------------------------------
global type Instance(object inst)
    return is_instance(inst)
end type

--------------------------------------------------------------------------------
global constant ATOM = routine_id("is_atom"),
                INSTANCE = routine_id("is_instance"),
                INTEGER = routine_id("is_integer"),
                OBJECT = routine_id("is_object"),
                METHOD = -999,
                SEQUENCE = routine_id("is_sequence")

--------------------------------------------------------------------------------

-- Local constants

--------------------------------------------------------------------------------
-- pointers within an Object
constant PROPERTY = 1,
         VALUE = 2,
         TYPE = 3,
         TYP_ERR = "**ERROR** Type conflict in Property '%s' when "

--------------------------------------------------------------------------------

-- Local variables

--------------------------------------------------------------------------------

sequence Classes,
         Objects

sequence ClassList

sequence className
sequence properties
sequence values
sequence types
integer objectCounter

--------------------------------------------------------------------------------

-- Initialisation

--------------------------------------------------------------------------------

objectCounter = 0
Classes = {}
ClassList = {}
Objects = {}

--------------------------------------------------------------------------------

-- Routines for Class creation

--------------------------------------------------------------------------------
-- Local routines
procedure AddClass()
    if length(ClassList) = 0 then
        ClassList = {className}
        Classes = {{properties,values,types}}
    else
        ClassList = append(ClassList,className)
        Classes = append(Classes,{properties,values,types})
    end if
end procedure

function CheckType(integer typ,object val)
    return call_func(typ,{val})
end function

function FindClass(sequence class)
    integer found
    found = find(class,ClassList)
    if found = 0 then
        printf(2,"\n**ERROR** Class: '%s' not found!\n",{class})
    end if
    return found
end function

function FindClassProperty(sequence class,sequence property)
    integer cl,
            found
    sequence props
    cl = FindClass(class)
    props = Classes[cl][PROPERTY]
    found = find(property,props)
    if found = 0 then
        printf(2,"\n**ERROR** Class %s: '%s' not found!\n",{{class},{property}})
    end if
    return found
end function

function FindObjectProperty(Instance self,sequence property)
    integer found
    sequence props
    props = Objects[self][PROPERTY]
    found = find(property,props)
    if found = 0 then
        printf(2,"\n**ERROR** Object #%d: '%s' not found!\n",{{self},{property}})
    end if
    return found
end function

function GetProperty(Instance self,sequence property)
    return Objects[self][VALUE][FindObjectProperty(self,property)]
end function

procedure SetProperty(Instance self,sequence property,object val)
    integer prop,
            typ
    prop = FindObjectProperty(self,property)
    typ = Objects[self][TYPE][prop]
    if typ = METHOD then
        printf(2,"\n**ERROR** Attempt made to change Method \"%s\". ",
               {Objects[self][PROPERTY][prop]})
        puts(2,"(Use \"Override\" function to do this).\n")
    else
        -- check type on each assignment
        if CheckType(typ,val) then
            Objects[self][VALUE][prop] = val
        else
            printf(2,TYP_ERR & "using \"Set\" Method!\n",{property})
        end if
    end if
end procedure

procedure Base()
    properties = {"Get","Set"}
    values = {routine_id("GetProperty"),routine_id("SetProperty")}
    types = {METHOD,METHOD}
end procedure

procedure report_error(sequence text)
    puts(2,text & " not defined!")
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
-- Class creation routines.

--------------------------------------------------------------------------------
--/topic Global routines
--/proc Class(sequence name)
-- Opens the class definition.

--------------------------------------------------------------------------------
global procedure Class(sequence name)
    className = name
    properties = {}
    values = {}
    types = {}
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
--/proc Extends(sequence parents)
-- Defines the Class' parent(s).

--------------------------------------------------------------------------------
global procedure Extends(sequence parents)
    integer found,
            parent
    sequence props
    sequence vals
    sequence typs
    if length(parents) = 0 then
        properties ={}
        values = {}
        types = {}
    elsif length(parents) >= 1 then
        parent = FindClass(parents[1])
        if parent > 0 then
            properties &= Classes[parent][PROPERTY]
            values &= Classes[parent][VALUE]
            types &= Classes[parent][TYPE]
        end if
    end if
    if length(parents) > 1 then
        -- to avoid possibility of same property/method defined twice
        for i = 2 to length(parents) do
            parent = FindClass(parents[i])
            if parent > 0 then
                props = Classes[parent][PROPERTY]
                vals = Classes[parent][VALUE]
                typs = Classes[parent][TYPE]
                for j = 1 to length(props) do
                    found = find(props[j],properties)
                    if found = 0 then
                        properties = append(properties,props[j])
                        values = append(values,vals[j])
                        types = append(types,typs[j])
                    end if
                end for
            end if
        end for
    end if
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
--/proc Property(sequence property,integer typ,object default)
-- Defines a new property, names it, declares its type, and specifies a default value.

--------------------------------------------------------------------------------
global procedure Property(sequence property,integer typ,object default)
    if CheckType(typ,default) then
        values = append(values,default)
    else
        if typ = SEQUENCE then
            values = append(values,EMPTY_STRING)
        else
            values = append(values,0)
        end if
        printf(2,TYP_ERR & "setting default!\n",{property})
    end if
    properties = append(properties,property)

    types = append(types,typ)
end procedure


--------------------------------------------------------------------------------
--/topic Global routines
--/proc Method(sequence method,integer rid)
-- Defines a new method, names it and indicates the routine-id to which it refers.

--------------------------------------------------------------------------------
global procedure Method(sequence method,integer rid)
    if rid > -1 then
        properties = append(properties,method)
        values = append(values,rid)
        types = append(types,METHOD)
    else
        printf(2,"\n\n**ERROR** Method '%s' undefined!\n",{method})
    end if
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
--/proc Override(sequence method,integer rid)
-- Defines the routine-id to which an existing method should now refer.

--------------------------------------------------------------------------------

global procedure Override(sequence method,integer rid)
    integer found
    found = find(method,properties)
    if found = 0 then
        printf(2,"\n**ERROR** Method: '%s' not found!\n",{method})
    elsif rid < 0 then
        puts(2,"\n**ERROR** Invalid routine-id!\n")
    else
        values[found] = rid
    end if
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
--/proc EndClass(sequence name)
-- Completes the class definition.

--------------------------------------------------------------------------------
global procedure EndClass(sequence name)
    if equal(name,className) then
        AddClass()
    else
        printf(2,"\n**ERROR** Class name conflict: '%s' is not '%s'!\n",
                 {name,className})
    end if
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
--/func MethodFunc(Instance self,sequence method,sequence args)
-- Invokes the <b>method</b> for the instance <b>self</b>, with the parameters <b>args</b>.
--/n Returns the value specified in the routine pointed to by the method.

--------------------------------------------------------------------------------
global function MethodFunc(Instance self,sequence method,sequence args)
    integer rid
    rid = GetProperty(self,method)
    if rid >= 0 then
        if length(args) = 0 then
            return call_func(rid,{self})
        else
            args = prepend(args,self)
            return call_func(rid,args)
        end if
    else
        report_error("Method")
    end if
end function

--------------------------------------------------------------------------------
--/topic Global routines
--/proc MethodProc(Instance self,sequence method,sequence args)
-- Invokes the <b>method</b> for the instance <b>self</b>, with the parameters <b>args</b>.

--------------------------------------------------------------------------------
global procedure MethodProc(Instance self,sequence method,sequence args)
    integer rid
    rid = GetProperty(self,method)
    if rid >= 0 then
        if length(args) = 0 then
            call_proc(rid,{self})
        else
            args = prepend(args,self)
            call_proc(rid,args)
        end if
    else
        report_error("Method")
    end if
end procedure

--------------------------------------------------------------------------------
--/topic Global routines
--/func New(sequence class,sequence args)
-- Creates an Instance of type class. See /Parms for details of parameters.
--/n The return value is a pointer to the Object.

--------------------------------------------------------------------------------
global function New(sequence class,sequence args)
    integer cl,
            found
    Base()
    cl = FindClass(class)
    properties &= Classes[cl][PROPERTY]
    values &= Classes[cl][VALUE]
    types &= Classes[cl][TYPE]
    -- set parameters
    if length(args) > 0 then
        for i = 1 to length(args) do
            found = find(args[i][PROPERTY],properties)
            if found > 0 then
                if CheckType(types[found],args[i][VALUE]) then
                    values[found] = args[i][VALUE]
                else
                    printf(2,TYP_ERR & "instantiating Object!\n",
                           {args[i][PROPERTY]})
                end if
            end if
        end for
    end if
    -- Add to Objects
    if objectCounter = 0 then
        Objects = {{properties,values,types}}
    else
        Objects = append(Objects,{properties,values,types})
    end if
    objectCounter += 1
    return objectCounter
end function

--------------------------------------------------------------------------------
--/topic Parms
-- Parameters for the /New function.
--/n /n The args should be sequences containing pairs:
--<ul>
--<li>name of property
--<li>value to be assigned to that property
--</ul>
-- For example:
--<pre>{{"TITLE","My title"},{"WIDTH,200}}</pre>

--------------------------------------------------------------------------------
--/topic Built-in Methods
-- Methods available for all Objects.

--------------------------------------------------------------------------------
--/topic Built-in Methods
--/func "Get"
--The return value is the value of the specified Property.
--/n /n Syntax: /MethodFunc(instance_name,"Get",{property_name})

--------------------------------------------------------------------------------
--/topic Built-in Methods
--/proc "Set"
-- Changes the value of the specified Property.
--/n /n Syntax: /MethodProc(instance_name,"Set",{property_name,value})
--/n If the value is not of the correct type then an error message is issued
--and the value remains unchanged.

--------------------------------------------------------------------------------
--/topic Inspection Routines
-- Feedback and testing outines.
--/n /n
-- These routines provide information about the defined Classes and instantiated
-- Objects.

--------------------------------------------------------------------------------
--/topic Inspection Routines
--/proc ListClasses()
-- Counts and lists the Classes defined.

--------------------------------------------------------------------------------
global procedure ListClasses()
    puts(1,"\n\n** List of Classes **")
    printf(1,"\n   At this point in time %d classes have been defined.",
           length(Classes))
    puts(1,"\n   These are: ")
    for i = 1 to length(Classes)-1 do
        puts(1,"\"" & ClassList[i] &"\"; ")
    end for
    puts(1,"\"" & ClassList[$] & "\".\n** End **\n\n")
end procedure

--------------------------------------------------------------------------------
--/topic Inspection Routines
--/proc ListClass(sequence class)
-- Details the contents of the specified Class.

--------------------------------------------------------------------------------
global procedure ListClass(sequence class)
    integer cl
    sequence props
    sequence vals
    sequence typs
    cl = FindClass(class)
    props = Classes[cl][PROPERTY]
    vals = Classes[cl][VALUE]
    typs = Classes[cl][TYPE]
    printf(1,"\n\n** Details of Class: \"%s\" **",{class})
    for i = 1 to length(props) do
        if typs[i] = METHOD then
            puts(1,"\n   Method")
        else
            puts(1,"\n   Property")
        end if
        printf(1," \"%s\":",{props[i]})
        if typs[i] = SEQUENCE then

            -- was a problem here with compound SEQUENCE:
            -- it crashed if SEQUENCE was compound;
            -- here is an incomplete "cure".
--            puts(1," value = {")
--            for x = 1 to length(vals[i]) do
--            	if sequence(vals[i][x]) then
--            		-- assume vals[i] is an ascii sequence
--            		-- (will still crash here if it is not; it
--            		--  needs a recursive solution on sequences)
--            		if x != length(vals[i]) then
--            			printf(1,"\"%s\", ",{vals[i][x]})
--            		else
--            			printf(1,"\"%s\"}",{vals[i][x]})
--            			? vals[i][x]
--            		end if
--            	else
--           		 if x != length(vals[i]) then
--           		 	printf(1, "%g, ", vals[i][x])
--           		 else
--           		 	printf(1, "%g}", vals[i][x])
--           		 end if
--            	end if
--            end for
			puts(1, " value = ")
			pretty_print(1,vals[i],"1")
			--? vals[i]
        else

            printf(1," value = %g",{vals[i]})

        end if
    end for
    puts(1,"\n** End **\n\n")
end procedure

--------------------------------------------------------------------------------
--/topic Inspection Routines
--/proc ListObject(Instance obj)
-- Details the current contents of the specified instance.

--------------------------------------------------------------------------------
global procedure ListObject(Instance obj)
    sequence props
    sequence vals
    sequence typs
    props = Objects[obj][PROPERTY]
    vals = Objects[obj][VALUE]
    typs = Objects[obj][TYPE]
    printf(1,"\n\n** Details of Object: %d **",{obj})
    for i = 1 to length(props) do
        if typs[i] = SEQUENCE then
            printf(1,"\n   Property \"%s\" has value \n",{props[i]})
            pretty_print(1,vals[i],"2")
            --? vals[i]
        else
            if typs[i] = METHOD then
                puts(1,"\n   Method")
            else
                puts(1,"\n   Property")
            end if
            printf(1," \"%s\" has value %g",{props[i],vals[i]})
        end if
    end for
    puts(1,"\n** End **\n")
end procedure